//version 1.00

// This file has been prepared for Doxygen automatic documentation generation.
/*! \file ********************************************************************
*
* Atmel Corporation
*
* File              : USI_TWI_Slave.c
* Compiler          : IAR EWAAVR 4.11A
* Revision          : $Revision: 1.14 $
* Date              : $Date: Friday, December 09, 2005 17:25:38 UTC $
* Updated by        : $Author: jtyssoe $
*
* Support mail      : avr@atmel.com
*
* Supported devices : All device with USI module can be used.
*
* AppNote           : AVR312 - Using the USI module as a I2C slave
*
* Description       : Functions for USI_TWI_receiver and USI_TWI_transmitter.
*
*
****************************************************************************/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include "USI_TWI_Slave.h"

extern unsigned char location,location2,location3;

/*! Static Variables
 */

static unsigned char TWI_slaveAddress;
volatile unsigned char USI_TWI_Overflow_State;
//static volatile unsigned char USI_TWI_Overflow_State;

/*! Local variables
 */
static uint8_t TWI_RxBuf[TWI_RX_BUFFER_SIZE];
static volatile uint8_t TWI_rx_buf_index = 0;



static uint8_t TWI_TxBuf[TWI_TX_BUFFER_SIZE];
static volatile uint8_t TWI_txDataInBuffer;

static volatile uint8_t busyTx, busyRx;


//********** USI_TWI functions **********//

/*! \brief
 * Initialise USI for TWI Slave mode.
 */
void USI_TWI_Slave_Initialise( unsigned char TWI_ownAddress )
{
	unsigned char i;

	busyTx = 0;
	busyRx = 0;

	for(i = 0; i < TWI_TX_BUFFER_SIZE; i++)
	{
		TWI_TxBuf[i] = 0;
	}


	TWI_slaveAddress = TWI_ownAddress;

    // Set SCL and SDA as output
    DDR_USI |= ( 1 << PORT_USI_SCL ) | ( 1 << PORT_USI_SDA );

	PORT_USI |=  (1<<PORT_USI_SCL);                               // Set SCL high
	PORT_USI |=  (1<<PORT_USI_SDA);                               // Set SDA high
	//DDR_USI  &=  ~(1<<PORT_USI_SCL);                              // Set SCL as input
	DDR_USI  &= ~(1<<PORT_USI_SDA);                               // Set SDA as input

	USICR  =  (1<<USISIE)|(0<<USIOIE)|                            // Enable Start Condition Interrupt. Disable Overflow Interrupt.
              (1<<USIWM1)|(0<<USIWM0)|                            // Set USI in Two-wire mode. No USI Counter overflow prior
                                                                  // to first Start Condition (potentail failure)
              (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
              (0<<USITC);
	USISR    = 0xF0;                                              // Clear all flags and reset overflow counter
}

/*
 * Call this function to set data that can be read by the master.
 * Include a pointer to the data to be sent if a SLA+W is received
 * The data is copied to the TWI TX buffer.
 * Include also how many bytes that should be sent.
 * The function will hold execution (loop) until the TWI has completed
 * a current SLA+W operation then initialize the next operation and return.
 *
 * Returns:	TWI_NO_ERROR if the update was successful.
 *			TWI_TIMEOUT_ERROR if the function timed out.
 */
unsigned char TWI_Transceiver_Update_Data( unsigned char *msg, unsigned char msgSize )
{
  	unsigned char i;
	
	while ( busyTx )// Wait until TWI is not in the middle of a TX.
	{
		if(USISR&(1<<USIPF) )//was a stop condition detected
		{
			busyTx = 0;
			busyRx = 0;//don't think it is needed but better safe then sorry
			//no need to clear it??
//			USISR |= (1<<USIPF);//causes TWI lockup.  Just USISR = _BV(USIPF) but not happy with
		}
		;
	}


    //Ensure this is run as an atomic block
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
 	{
  		if (msgSize <= TWI_TX_BUFFER_SIZE)
  		{
   	 		TWI_txDataInBuffer = msgSize;        		// Number of data to transmit.
			TWI_txDataInBuffer+=1;	//darrell: add one byte for timestamp
  		}
  		else
  		{
   			TWI_txDataInBuffer = TWI_TX_BUFFER_SIZE;	//Only copy what we can
  		}


  		for ( i = 0; i < (TWI_txDataInBuffer-1); i++ )      // Copy data that may be transmitted if the TWI Master requests data.
    	{
    		TWI_TxBuf[ i ] = msg[ i ];
    	}

	}

    return TWI_NO_ERROR;
}



/*
 * Call this function to read any received data from the TWI RX buffer.
 *
 * Returns: The number of bytes received.
 * 			Note if the TWI is in the middle of a receive it will return 0
 */
unsigned char TWI_Get_Data_From_Transceiver(unsigned char *msg,unsigned char msgSize)
{
	unsigned char i = 0;
	unsigned char temp_SREG;

	//Ensure this is run as an atomic block
	temp_SREG = SREG;
	cli();

	if(busyRx) {  // Is the TWI in the middle of a RX.
		if(USISR&(1<<USIPF))  {// was a stop condition detected
			busyRx = 0;
			busyTx = 0;// don't think it is needed but better safe then sorry
			// could clear it, but not required as will be cleared at next start.
			//USISR = _BV(USIPF);//(1<<USIPF);//clear the stop flag.  Old way caused errors
			}
		else  {
			SREG = temp_SREG; //restore interrupts if enabled
			return 0;
			}
		}

   	if(TWI_rx_buf_index==0) {
		SREG = temp_SREG; //restore interrupts if enabled
    	return 0;
    	}

	while(TWI_rx_buf_index&&(i<msgSize)) {
		msg[i] = TWI_RxBuf[i];
    	i++;
    	TWI_rx_buf_index--;
		}

    TWI_rx_buf_index = 0;

	SREG = temp_SREG; // restore interrupts if enabled
    return i;
}


/* Usi start condition ISR
 * Detects the USI_TWI Start Condition and intialises the USI
 * for reception of the "TWI Address" packet.
 */
ISR(USI_START_VECTOR)
{
   // unsigned char tmpUSISR;                                         // Temporary variable to store volatile
   // tmpUSISR = USISR;                                               // Not necessary, but prevents warnings

    // Set default starting conditions for new TWI package
    USI_TWI_Overflow_State = USI_SLAVE_CHECK_ADDRESS;
    DDR_USI  &= ~(1<<PORT_USI_SDA);                                 // Set SDA as input

   // while ((PIN_USI & (1<<PORT_USI_SCL)) & !(tmpUSISR & (1<<USIPF)));   // Wait for SCL to go low to ensure the "Start Condition" has completed.
                                                                       // If a Stop condition arises then leave the interrupt to prevent waiting forever.

//    while((PIN_USI&(1<<PORT_USI_SCL))&!(USISR&(1<<USIPF)));   // Wait for SCL to go low to ensure the "Start Condition" has completed.
    while((PIN_USI&(1<<PORT_USI_SCL))&&!(PIN_USI & (1<<PORT_USI_SDA)) );   // Wait for SCL to go low to ensure the "Start Condition" has completed.
	                                                          // If a Stop condition arises then leave the interrupt to prevent waiting forever.

	if (PIN_USI & (1<<PORT_USI_SDA))                                // STOP 
       USICR  =  (1<<USISIE)|(0<<USIOIE)|                            // Enable Overflow and Start Condition Interrupt. (Keep StartCondInt to detect RESTART)
			     (1<<USIWM1)|(0<<USIWM0)|                            // Set USI in Two-wire mode.
                 (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
                 (0<<USITC);
	else  
       USICR  =  (1<<USISIE)|(1<<USIOIE)|                            // Enable Overflow and Start Condition Interrupt. (Keep StartCondInt to detect RESTART)
			     (1<<USIWM1)|(1<<USIWM0)|                            // Set USI in Two-wire mode.
                 (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
                 (0<<USITC);


 /*      USICR  =  (1<<USISIE)|(1<<USIOIE)|                            // Enable Overflow and Start Condition Interrupt. (Keep StartCondInt to detect RESTART)
        //      (1<<USIWM1)|(0<<USIWM0)|   //darrell changed to disable SCL hold low
			    (1<<USIWM1)|(1<<USIWM0)|                            // Set USI in Two-wire mode.
                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
                (0<<USITC);
*/
  //  USISR  =    (1<<USI_START_COND_INT)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      // Clear flags
  //              (0x0<<USICNT0);                                     // Set USI to sample 8 bits i.e. count 16 external pin toggles.
																	//this is read only (1<<USIDC)
	USISR  =    (1<<USI_START_COND_INT)|(1<<USIOIF)|(1<<USIPF)|      // Clear flags
                (0x0<<USICNT0);

	busyTx = 0;
	busyRx = 0;
}


/*! \brief USI counter overflow ISR
 * Handles all the comunication. Is disabled only when waiting
 * for new Start Condition.
 */
ISR(USI_OVERFLOW_VECTOR)
{
	static unsigned char TWI_bufPtr = 0;

				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);

				//DDRB |= (1<<PINB0);
				//PORTB &= ~(1<<PINB0);

	switch (USI_TWI_Overflow_State)
	{	// Copy data from USIDR and send ACK. Next USI_SLAVE_REQUEST_DATA
    	

    	// ---------- Address mode ----------
    	// Check address and send ACK (and next USI_SLAVE_SEND_DATA) if OK, else reset USI.
    	case USI_SLAVE_CHECK_ADDRESS:
      		if((USIDR==0) || ((USIDR>>1) == TWI_slaveAddress))  //general call ack
      		{
        		if(USIDR & 0x01 )//I2C read operation
				{
          			USI_TWI_Overflow_State = USI_SLAVE_SEND_DATA;
		  			busyTx = 1;
		  			TWI_bufPtr = 0;
        		}
				else  //write operation
				{
          			USI_TWI_Overflow_State = USI_SLAVE_REQUEST_DATA;
		  			busyRx = 1;
		  			TWI_rx_buf_index = 0;
        		}

		  		SET_USI_TO_SEND_ACK();
				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);
				//location = 0x01;
				//location2 = 0;
      		}
      		else
      		{
        		SET_USI_TO_TWI_START_CONDITION_MODE();
      		}

			break;



    		// ----- Master write data mode ------
    		// Check reply and goto USI_SLAVE_SEND_DATA if OK, else reset USI.
    	case USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA:

			if ( USIDR ) // If NACK, the master does not want more data.
      		{
				busyTx = 0;

				SET_USI_TO_TWI_START_CONDITION_MODE();
        		return;
      		}

		// From here we just drop straight into USI_SLAVE_SEND_DATA if the master sent an ACK

    	// Copy data from buffer to USIDR and set USI to shift byte. Next USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA
    	case USI_SLAVE_SEND_DATA:
			
			//darrell: add timestamp at time of send
		//	TWI_TxBuf[ (TWI_txDataInBuffer-1)] = TCNT1;	//add the time to end of message


			if ( TWI_bufPtr < TWI_txDataInBuffer )
			{
				USIDR = TWI_TxBuf[TWI_bufPtr++];
			}
    		else // If the buffer is empty then:
    		{
				busyTx = 0;
       			SET_USI_TO_TWI_START_CONDITION_MODE();
        		return;
			}

			USI_TWI_Overflow_State = USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA;//??not sure why not just checking ack??
    		SET_USI_TO_SEND_DATA();
    		break;

    	// Set USI to sample reply from master. Next USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA
    	case USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA:

	  		USI_TWI_Overflow_State = USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA;
      		SET_USI_TO_READ_ACK();
      		break;

    	// ----- Master read data mode ------
    	// Set USI to sample data from master. Next USI_SLAVE_GET_DATA_AND_SEND_ACK.
		case USI_SLAVE_REQUEST_DATA:

	  		USI_TWI_Overflow_State = USI_SLAVE_GET_DATA_AND_SEND_ACK;
				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);
      		SET_USI_TO_READ_DATA();
				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);
      		break;

    	case USI_SLAVE_GET_DATA_AND_SEND_ACK:

			TWI_RxBuf[TWI_rx_buf_index++] = USIDR;//here or in an else below
			location2 = USIDR;
	  		//Ensure that our buffer has enough room for more data
			/*if(TWI_rx_buf_index >= TWI_RX_BUFFER_SIZE)//TWI_txDataInBuffer
			{
				busyRx = 0;
       			SET_USI_TO_TWI_START_CONDITION_MODE();//now more room for data so stop recieving
				//location <<= 1;
			}
			else {
      		*/	USI_TWI_Overflow_State = USI_SLAVE_REQUEST_DATA;
				//location2 = 6;
				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);
      			SET_USI_TO_SEND_ACK();
				//PORTB |= (1<<PINB1);
				//PORTB &= ~(1<<PINB1);
			//	}
      		break;
	}

}



